package edu.rice.comp211.laundry.list;


import edu.rice.comp211.laundry.iterator.BiIteratorI;
import edu.rice.comp211.laundry.iterator.IteratorI;

/**
 * Singleton class holding the utility methods on BiLists used in the Laundry Simulation.
 */
public class BiListUtil {
	/**
	 * Singleton instance
	 */
	public static final BiListUtil ONLY = new BiListUtil();
	private BiListUtil() {}


	/**
	 * Deletes the first occurrence of elt from host.
	 * @param <T> the type of the element being deleted -- inferred from host and elt.
	 * @param host the list to delete from 
	 * @param elt the element to delete
	 * @return true if the element was deleted, false otherwise
	 */
	public <T> boolean delete(BiListI<T> host, T elt) {
		for (IteratorI<T> i = host.newIterator(); !i.atEnd(); i.next()) {
			if (i.currentItem().equals(elt)) {
				i.remove();
				return true;
			}
		}
		return false;
	}

	/**
	 * Deletes the first occurrence of elt in each list (of type BiList<I>) in host.
	 * @param <T> the type of element that is being deleted -- inferred from host and elt
	 * @param host the list to delete from 
	 * @param elt the element to delete
	 * @return true if the element was deleted, false otherwise
	 */
	public <T> boolean mapDelete(BiListI<BiListI<T>> host, T elt) {
		// requires that this is CircList of CircList
		for (IteratorI<BiListI<T>> i = host.newIterator(); !i.atEnd(); i.next()) {
			if (delete(i.currentItem(), elt)) {
				return true;
			}
		}
		return false;
	}


	/**
	 * Delete the first occurrence of an element for which the given Predicate returns true.  
	 * The list is searched from the front to the rear.
	 * Returns the the removed element or the given nullValue if not found. 
	 * @param <T> The type of element to be removed -- inferred from host, pred and nullValue
	 * @param host the list to delete from 
	 * @param pred the Predicate to use
	 * @param nullValue the value used to represent not finding an element to delete.
	 * @return the removed element or nullValue if nothing found.
	 */
	public <T> T deleteMatchFront(BiListI<T> host, Predicate<T> pred, T nullValue) {
		for (IteratorI<T> i = host.newIterator(); !i.atEnd(); i.next()) {
			if (pred.test(i.currentItem())) {
				T res = i.currentItem();
				i.remove();
				return res;
			}
		}
		return nullValue;
	}

	/**
	 * Delete the first occurrence of an element for which the given Predicate returns true.  
	 * The list is searched from the rear to the front.
	 * Returns the the removed element or the given nullValue if not found. 
	 * @param <T> The type of element to be removed -- inferred from host, pred and nullValue
	 * @param host the list to delete from 
	 * @param pred the Predicate to use
	 * @param nullValue the value used to represent not finding an element to delete.
	 * @return the removed element or nullValue if nothing found.
	 */
	public <T> T deleteMatchRear(BiListI<T> host, Predicate<T> pred, T nullValue) {
		BiIteratorI<T> i = host.newIterator();
		for (i.last(); !i.atEnd(); i.prev()) {
			if (pred.test(i.currentItem())) {
				T res = i.currentItem();
				i.remove();
				return res;
			}
		}
		return nullValue;
	}


	/**
	 * Removes the first size elements from host and returns them as the result. If size elements are not available, it
	 * removes and returns the elements that are available.
	 * @param <T>  the type of elements to remove -- inferred from host
	 * @param host the list to remove from
	 * @param size the maximum number of elements to remove
	 * @return a BiList of the removed elements, in their original order
	 */
	public <T> BiListI<T> remListFront(BiListI<T> host, int size) {
		BiListI<T> result = new BiList<T>();
		IteratorI<T> i = host.newIterator();
		while (!i.atEnd() && size > 0) {
			size--;
			// i.remove mutates this
			result.insertRear(i.remove());
		}
		return result;
	}


	/**
	 * Generates a string representation for host with items separated by ", " and no leading or trailing text.
	 * @param <T> the type of elements being processed -- inferred from host
	 * @param host the list to process
	 * @return the string representation of the list.
	 */
	public <T> String toStringWithCommas(BiListI<T> host) {
		String result = "";
		BiIteratorI<T> it = host.newIterator();
		if (!it.atEnd()) {
			/* Convert first item to string with no prefix. */
			result += it.currentItem();
			it.next();
			/* Convert remaining items to strings with ", " prefix. */
			while (!it.atEnd()) {
				result += ", " +it.currentItem();
				it.next();
			}
		}
		return result;
	}

}
