package edu.rice.comp211.laundry.ui;

import edu.rice.comp211.laundry.command.*;
import edu.rice.comp211.laundry.garment.Garment;
import edu.rice.comp211.laundry.garment.Pants;
import edu.rice.comp211.laundry.garment.Shirt;
import edu.rice.comp211.laundry.garment.Socks;
import edu.rice.comp211.laundry.io.IOProcess;
import edu.rice.comp211.laundry.io.TerminalIO;
import edu.rice.comp211.laundry.list.BiListI;
import edu.rice.comp211.laundry.student.Student;
import edu.rice.comp211.laundry.student.StudentEnvironment;
import edu.rice.comp211.laundry.sync.SyncBiList;

import javax.swing.*;
import javax.swing.border.LineBorder;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.*;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.*;
import java.util.List;

/**
 * Date: Mar 23, 2010
 * Time: 11:35:39 PM
 */
public class LaundryController {

   private final JFrame frame;
   private final ResourceBundle bundle;

   private StudentModel studentModel;
   private IOProcess ioProcess;
   private List<Command> commandList;
   private Map<JList, String> garmentListMap;
   private List<JComponent> actionElements;

   private JPanel framePanel;
   private JLabel statusLabel;
   private JButton soloNextCommandButton;
   private JButton soloRunAllButton;
   private JButton threadedNextCommandButton;
   private JButton threadedRunAllButton;
   private JPanel fileInputButtonsPanel;
   private JPanel uiInputButtonsPanel;
   private JPanel threadedInputButtonsPanel;
   private JComboBox inputModeCombo;
   private JPanel mBottomPanel;

   LaundryController(final JFrame frame) {
      this.frame = frame;
      this.bundle = Application.getInstance().getResourceBundle();
      this.garmentListMap = new HashMap<JList, String>();
      this.actionElements = new LinkedList<JComponent>();

      createModel(1, (String) null);
      createUI();
      createMenu();

      frame.pack();
   }

   // Tries to load the model. If that fails a new model is returned
   private void createModel(final int studentCount, final String... pInputFileName) {
      if (studentCount != pInputFileName.length) {
         throw new RuntimeException("Student count and file names do not match");
      }
      if (studentModel == null) {
         studentModel = new StudentModel();
      } else {
         studentModel.clearStudents();
      }

      if (commandList == null) {
         commandList = new LinkedList<Command>();
      } else {
         commandList.clear();
      }

      if (pInputFileName[0] != null) {
         ioProcess = new TerminalIO(pInputFileName[0]);
      } else {
         // TODO Dummy IOProcess for Student. Discuss refactoring Student to accept OutputStream instead of IOProcess
         ioProcess = new IOProcess() {
            public PrintStream open(final StudentEnvironment s, final boolean debug) {
               return System.out;
            }

            public Command nextCommand() throws IOException {
               throw new RuntimeException("Unexpected call on nextCommand");
            }
         };
      }
      studentModel.addStudent(createNewStudent(0, studentCount > 1), ioProcess);
      for (int i = 1; i < studentCount; i++) {
         studentModel.addStudent(createNewStudent(i, true), new TerminalIO(pInputFileName[i]));
      }
   }

   private Student createNewStudent(final int index, final boolean syncMode) {
      final Shirt defaultShirt = new Shirt(bundle.getString("default." + index + ".shirt"));
      final Pants defaultPants = new Pants(bundle.getString("default." + index + ".pants"));
      final Socks defaultSocks = new Socks(bundle.getString("default." + index + ".socks"));
      final String studentName = bundle.getString("studentName." + index);
      if (index == 0) {
         if (syncMode) {
            return new Student(studentName, defaultShirt, defaultPants, defaultSocks,
                  new SyncBiList<Garment>(), new SyncBiList<Garment>(), new SyncBiList<Garment>(),
                  new SyncBiList<Garment>(), new SyncBiList<BiListI<Garment>>());
         } else {
            return new Student(studentName, defaultShirt, defaultPants, defaultSocks);
         }
      } else {
         // Students must share piles
         final Student firstStudent = studentModel.firstStudent();
         return new Student(studentName, defaultShirt, defaultPants, defaultSocks,
               firstStudent.cleanShirts(), firstStudent.cleanPants(), firstStudent.cleanSocks(),
               firstStudent.dirtyPile(), firstStudent.laundryRoom());
      }
   }

   private void createUI() {
      frame.setLayout(new BorderLayout());

      framePanel = new JPanel(new BorderLayout());
      final JPanel garmentListPanel = createGarmentListPanel();
      framePanel.add(garmentListPanel, BorderLayout.CENTER);

      mBottomPanel = new JPanel(new BorderLayout(3, 3));

      final JPanel westPanel = new JPanel(new GridLayout(2, 1));
      westPanel.add(new JLabel(bundle.getString("label.inputMode")));
      inputModeCombo = new JComboBox(InputMode.values());
      inputModeCombo.addActionListener(new DynamicAction(this, "inputModeChanged"));
      actionElements.add(inputModeCombo);
      westPanel.add(inputModeCombo);
      mBottomPanel.add(westPanel, BorderLayout.WEST);

      fileInputButtonsPanel = createFileInputButtonsPanel();
      mBottomPanel.add(fileInputButtonsPanel, BorderLayout.CENTER);

      uiInputButtonsPanel = createUiInputButtonsPanel();

      threadedInputButtonsPanel = createThreadedInputButtonsPanel();

      statusLabel = new JLabel("Program Started with input file: testInput.txt", JLabel.CENTER);
      statusLabel.setFont(new Font(Font.SERIF, Font.PLAIN, 12));
      mBottomPanel.add(statusLabel, BorderLayout.SOUTH);

      framePanel.add(mBottomPanel, BorderLayout.SOUTH);

      frame.add(framePanel, BorderLayout.CENTER);
   }

   private JPanel createFileInputButtonsPanel() {
      final JPanel buttonsPanel = new JPanel(new GridLayout(2, 5));
      final JButton newInputButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(newInputButton, bundle.getString("button.newInput"));
      newInputButton.addActionListener(new DynamicAction(this, "openNewFile"));
      buttonsPanel.add(newInputButton);
      actionElements.add(newInputButton);

      soloNextCommandButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(soloNextCommandButton, bundle.getString("button.next"));
      soloNextCommandButton.addActionListener(new DynamicAction(this, "nextCommandInSoloMode"));
      buttonsPanel.add(soloNextCommandButton);
      soloNextCommandButton.setEnabled(false);
      actionElements.add(soloNextCommandButton);

      soloRunAllButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(soloRunAllButton, bundle.getString("button.runAll"));
      soloRunAllButton.addActionListener(new DynamicAction(this, "runInSoloModeAllRemainingCommands"));
      buttonsPanel.add(soloRunAllButton);
      soloRunAllButton.setEnabled(false);
      actionElements.add(soloRunAllButton);

      final JButton viewExecCmdsButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(viewExecCmdsButton, bundle.getString("button.viewExecCommands"));
      viewExecCmdsButton.addActionListener(new DynamicAction(this, "viewExecutedCommands"));
      buttonsPanel.add(viewExecCmdsButton);
      actionElements.add(viewExecCmdsButton);

      final JButton saveCommandsButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(saveCommandsButton, bundle.getString("button.saveCommands"));
      saveCommandsButton.addActionListener(new DynamicAction(this, "saveExecutedCommands"));
      buttonsPanel.add(saveCommandsButton);
      actionElements.add(saveCommandsButton);
      
      final JButton exitButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(exitButton, bundle.getString("button.exit"));
      exitButton.addActionListener(new DynamicAction(Application.getInstance(), "exit"));
      buttonsPanel.add(exitButton);
      actionElements.add(exitButton);

      return buttonsPanel;
   }

   private JPanel createThreadedInputButtonsPanel() {
      final JPanel buttonsPanel = new JPanel(new GridLayout(2, 3));
      final JButton newInputButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(newInputButton, bundle.getString("button.newInput"));
      newInputButton.addActionListener(new DynamicAction(this, "openThreadedNewFile"));
      buttonsPanel.add(newInputButton);
      actionElements.add(newInputButton);

      threadedNextCommandButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(threadedNextCommandButton, bundle.getString("button.next"));
      threadedNextCommandButton.addActionListener(new DynamicAction(this, "nextCommandInThreadedMode"));
      buttonsPanel.add(threadedNextCommandButton);
      threadedNextCommandButton.setEnabled(false);
      actionElements.add(threadedNextCommandButton);

      threadedRunAllButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(threadedRunAllButton, bundle.getString("button.runAll"));
      threadedRunAllButton.addActionListener(new DynamicAction(this, "runInThreadedModeAllRemainingCommands"));
      buttonsPanel.add(threadedRunAllButton);
      threadedRunAllButton.setEnabled(false);
      actionElements.add(threadedRunAllButton);

      final JButton viewExecCmdsButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(viewExecCmdsButton, bundle.getString("button.viewExecCommands"));
      viewExecCmdsButton.addActionListener(new DynamicAction(this, "viewExecutedCommands"));
      buttonsPanel.add(viewExecCmdsButton);
      actionElements.add(viewExecCmdsButton);

      final JButton exitButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(exitButton, bundle.getString("button.exit"));
      exitButton.addActionListener(new DynamicAction(Application.getInstance(), "exit"));
      buttonsPanel.add(exitButton);
      actionElements.add(exitButton);

      return buttonsPanel;
   }

   private JPanel createGarmentListPanel() {
      final JPanel garmentListPanel = new JPanel(new FlowLayout());

      final JPanel closetPanel = new JPanel(new GridLayout(1, 3));

      final JList cleanShirtList = createGarmentJList("cleanShirts");
      final JPanel cleanShirtPanel = createGarmentListPanel(cleanShirtList, "label.closetShirt");
      closetPanel.add(cleanShirtPanel);

      final JList cleanPantsList = createGarmentJList("cleanPants");
      final JPanel cleanPantsPanel = createGarmentListPanel(cleanPantsList, "label.closetPants");
      closetPanel.add(cleanPantsPanel);

      final JList cleanSocksList = createGarmentJList("cleanSocks");
      final JPanel cleanSocksPanel = createGarmentListPanel(cleanSocksList, "label.closetSocks");
      closetPanel.add(cleanSocksPanel);

      final JPanel closetContainer = new JPanel(new BorderLayout(2, 2));
      final JLabel closetLabel = new JLabel(bundle.getString("label.closetShelf"), JLabel.CENTER);
      closetLabel.setFont(new Font(Font.SERIF, Font.BOLD, 20));
      closetLabel.setForeground(Color.black);
      closetContainer.add(closetLabel, BorderLayout.NORTH);
      closetContainer.add(closetPanel, BorderLayout.CENTER);
      garmentListPanel.add(closetContainer);

      final JList currentDressList = createGarmentJList("currentGarments");
      final JPanel currentDressPanel = createGarmentListPanel(currentDressList, "label.currentDress");
      garmentListPanel.add(currentDressPanel);

      final JList dirtyLaundryList = createGarmentJList("dirtyPile");
      final JPanel dirtyLaundryPanel = createGarmentListPanel(dirtyLaundryList, "label.dirtyLaundryPile");
      garmentListPanel.add(dirtyLaundryPanel);

      final JList laundryRoomList = createGarmentJList("laundryRoom");
      final JPanel laundryRoomPanel = createGarmentListPanel(laundryRoomList, "label.laundryRoomTable");
      garmentListPanel.add(laundryRoomPanel);
      return garmentListPanel;
   }

   private JPanel createGarmentListPanel(final JList pJList, final String pTitleKey, final Object... pArgs) {

      final JPanel containerPanel = new JPanel(new BorderLayout(2, 2));

      final JLabel titleLabel = new JLabel(bundle.getString(pTitleKey), JLabel.CENTER);
      titleLabel.setFont(new Font(Font.SERIF, Font.BOLD, 20));
      titleLabel.setForeground(Color.black);
      containerPanel.add(titleLabel, BorderLayout.NORTH);

      final JScrollPane scrollPane = new JScrollPane(pJList);
      scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
      scrollPane.setPreferredSize(new Dimension(150, 550));
      containerPanel.add(scrollPane, BorderLayout.CENTER);

      return containerPanel;
   }

   private JList createGarmentJList(final String modelMethod) {
      final JList jList = new JList();
      jList.setBorder(new LineBorder(Color.black, 2));
      jList.setCellRenderer(new LaundryListCellRenderer());
      garmentListMap.put(jList, modelMethod);
      return jList;
   }

   // Creates and populates the menu for the app
   private void createMenu() {
      // NOTE: notice that many of the actions in here are null, they will be filled in later on.
      final JMenuBar menuBar = new JMenuBar();
      frame.setJMenuBar(menuBar);

      final JMenu fileMenu = MnemonicHelper.createMenu(bundle.getString("menu.file"));
      menuBar.add(fileMenu);
      actionElements.add(fileMenu);
      final JMenuItem exitMI = MnemonicHelper.createMenuItem(fileMenu, bundle.getString("menu.exit"));
      exitMI.addActionListener(new DynamicAction(Application.getInstance(), "exit"));
      actionElements.add(exitMI);

      final JMenu helpMenu = MnemonicHelper.createMenu(bundle.getString("menu.help"));
      menuBar.add(helpMenu);
      actionElements.add(helpMenu);
      createMenuItem(helpMenu, "menu.about", new DynamicAction(this, "showAbout"));
   }

   // Conveniance to create a configure a JMenuItem.
   private JMenuItem createMenuItem(JMenu menu, String key, Action action) {
      final JMenuItem jMenuItem = (action != null) ? new JMenuItem(action) : new JMenuItem();
      MnemonicHelper.configureTextAndMnemonic(jMenuItem, Application.getResourceAsString(key));
      menu.add(jMenuItem);
      return jMenuItem;
   }

   public void showAbout() {
      AboutBox.getInstance().show(SwingUtilities.getWindowAncestor(frame));
   }

   public void nextCommandInSoloMode() {
      executeCommand();
   }

   private boolean executeCommand() {
      try {
         final Command command = ioProcess.nextCommand();
         if (executeCommand(command)) {
            return false;
         }
         return true;
      } catch (IOException ex) {
         ex.printStackTrace();
         setStatus("message.exception", ex.getMessage());
      }
      return false;
   }

   private boolean executeCommand(final Command pCommand) {
      if (pCommand == null) {
         enableCommandButtons(false);
         return true;
      }
      // Add to the list of executed commands
      commandList.add(pCommand);

      setStatus("message.processing", pCommand.toString());
      studentModel.executeCommand(pCommand);
      setStatus("message.processed", pCommand.toString());
      updateJListModels();
      return false;
   }

   public void runInSoloModeAllRemainingCommands() {
      enableComponents(this.actionElements, false);
      if (executeCommand()) {
         updateAndInvokeLater(new Runnable() {
            public void run() {
               runInSoloModeAllRemainingCommands();
            }
         });
      } else {
         enableComponents(this.actionElements, true);
         enableCommandButtons(false);
         setStatus("message.finished", null);
      }
   }

   public void nextCommandInThreadedMode() {
      nextCommandInThreadedMode(true);
   }

   private void nextCommandInThreadedMode(final boolean makeInitialCall) {
      enableComponents(this.actionElements, false);
      if (makeInitialCall) {
         studentModel.runOneStep();
      }
      if (!studentModel.doneExecuting()) {
         System.out.println("nextCommandInThreadedMode: Invoking next step");
         updateAndInvokeLater(new Runnable() {
            public void run() {
               nextCommandInThreadedMode(false);
            }
         });
      } else {
         System.out.println("nextCommandInThreadedMode: done with next step");
         updateJListModels();
         framePanel.repaint();
         enableComponents(this.actionElements, true);
         enableCommandButtons(true);
      }
   }

   public void runInThreadedModeAllRemainingCommands() {
      runInThreadedModeAllRemainingCommands(true);
   }

   private void runInThreadedModeAllRemainingCommands(final boolean makeInitialCall) {
      enableComponents(this.actionElements, false);
      if (makeInitialCall) {
         studentModel.runAllSteps();
      }
      if (!studentModel.doneExecuting()) {
         updateAndInvokeLater(new Runnable() {
            public void run() {
               runInThreadedModeAllRemainingCommands(false);
            }
         });
      } else {
         updateJListModels();
         framePanel.repaint();
         enableComponents(this.actionElements, true);
         enableCommandButtons(false);
         setStatus("message.finished", null);
      }
   }

   private void updateAndInvokeLater(final Runnable pRunnable) {
      try {
         updateJListModels();
         framePanel.repaint();
         SwingUtilities.invokeLater(pRunnable);
         Thread.sleep(Integer.parseInt(bundle.getString("animationSpeed")));
      } catch (Exception e) {
         e.printStackTrace();
      }
   }

   public void openNewFile() {
      final JFileChooser fc = new JFileChooser(new File(System.getProperty("user.dir"))); // Opens dialog in current working directory
      fc.setFileSelectionMode(JFileChooser.FILES_ONLY);

      int returnVal = fc.showDialog(frame, bundle.getString("fileChooser.title"));
      if (returnVal == JFileChooser.APPROVE_OPTION) {
         final File file = fc.getSelectedFile();
         //This is where a real application would open the file.
         setStatus("message.opening", file.getAbsolutePath());
         createModel(1, file.getAbsolutePath());
         updateJListModels();
         enableCommandButtons(true);
         setStatus("message.ready", file.getAbsolutePath());
      } else {
         setStatus("message.openCancelled", null);
      }
   }

   public void viewExecutedCommands() {
      final JDialog jDialog = new JDialog(frame, true);

      final JList jList = new JList(studentModel.getExecCommands().toArray());
      jList.setSelectionMode(ListSelectionModel.SINGLE_INTERVAL_SELECTION);
      jList.setLayoutOrientation(JList.HORIZONTAL_WRAP);
      jList.setVisibleRowCount(-1);

      final JScrollPane listScroller = new JScrollPane(jList);
      listScroller.setPreferredSize(new Dimension(300, 200));
      jDialog.add(listScroller);      

      jDialog.pack();
      jDialog.setLocationRelativeTo(frame);
      jDialog.setVisible(true);
   }

   public void openThreadedNewFile() {
      final JDialog jDialog = new JDialog(frame, true);

      final int maxStudents = 4;
      jDialog.setLayout(new GridLayout(maxStudents + 2, 1));

      jDialog.add(new JLabel(bundle.getString("message.chooseAtLeastTwoFiles")));

      final JTextField[] textFields = new JTextField[maxStudents];
      for (int i = 0; i < maxStudents; i++) {
         final JPanel loopPanel = new JPanel(new BorderLayout());
         final JTextField loopTextfield = new JTextField(80);
         textFields[i] = loopTextfield;
         loopPanel.add(textFields[i], BorderLayout.CENTER);

         final JButton fileChooserButton = new JButton(bundle.getString("button.chooseFile"));
         fileChooserButton.addActionListener(new ActionListener() {
            public void actionPerformed(final ActionEvent e) {
               final JFileChooser fc = new JFileChooser(new File(System.getProperty("user.dir")));
               fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
               int returnVal = fc.showDialog(frame, bundle.getString("fileChooser.title"));
               if (returnVal == JFileChooser.APPROVE_OPTION) {
                  final File file = fc.getSelectedFile();
                  final String filePath = file.getAbsolutePath();
                  loopTextfield.setText(filePath);
                  loopTextfield.setCaretPosition(filePath.length());
               }
            }
         });
         loopPanel.add(fileChooserButton, BorderLayout.EAST);

         jDialog.add(loopPanel);
      }
      final JPanel buttonPanel = new JPanel(new BorderLayout());
      final JButton closeButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(closeButton, bundle.getString("button.create"));
      closeButton.addActionListener(new ActionListener() {
         public void actionPerformed(final ActionEvent e) {
            jDialog.setVisible(false);
         }
      });
      buttonPanel.add(closeButton, BorderLayout.EAST);
      jDialog.add(buttonPanel);

      jDialog.pack();
      jDialog.setLocationRelativeTo(frame);
      jDialog.setVisible(true);

      final List<String> fileList = new ArrayList<String>(maxStudents);
      for (int i = 0; i < maxStudents; i++) {
         final String loopFilePath = textFields[i].getText();
         if (loopFilePath.length() > 0) {
            fileList.add(loopFilePath);
         }
      }
      if (fileList.size() > 1) {
         createModel(fileList.size(), fileList.toArray(new String[fileList.size()]));
         updateJListModels();
         enableCommandButtons(true);
         setStatus("message.threadedReady", String.valueOf(fileList.size()) + " " + bundle.getString("message.students"));
      }
   }

   public void inputModeChanged() {
      final Object selectedItem = inputModeCombo.getSelectedItem();
      if (InputMode.File.equals(selectedItem)) {
         createModel(1, (String) null);
         mBottomPanel.remove(uiInputButtonsPanel);
         mBottomPanel.remove(threadedInputButtonsPanel);
         enableCommandButtons(false);
         mBottomPanel.add(fileInputButtonsPanel, BorderLayout.CENTER);
      } else if (InputMode.GUI.equals(selectedItem)) {
         createModel(1, (String) null);
         mBottomPanel.remove(fileInputButtonsPanel);
         mBottomPanel.remove(threadedInputButtonsPanel);
         mBottomPanel.add(uiInputButtonsPanel, BorderLayout.CENTER);
      } else if (InputMode.Threaded.equals(selectedItem)) {
         createModel(2, null, null);
         mBottomPanel.remove(fileInputButtonsPanel);
         mBottomPanel.remove(uiInputButtonsPanel);
         enableCommandButtons(false);
         mBottomPanel.add(threadedInputButtonsPanel, BorderLayout.CENTER);
      }
      setStatus("message.inputMode", selectedItem.toString());
      updateJListModels();
   }

   private void enableCommandButtons(boolean state) {
      soloNextCommandButton.setEnabled(state);
      soloRunAllButton.setEnabled(state);
      threadedNextCommandButton.setEnabled(state);
      threadedRunAllButton.setEnabled(state);
   }

   private void setStatus(final String messageKey, final String suffix) {
      final String bundleMessage = bundle.getString(messageKey);
      if (suffix != null) {
         statusLabel.setText(bundleMessage + suffix);
      } else {
         statusLabel.setText(bundleMessage);
      }
   }

   private void updateJListModels() {
      for (final Map.Entry<JList, String> loopEntry : garmentListMap.entrySet()) {
         updateJListModel(loopEntry.getKey(), loopEntry.getValue());
      }
      frame.repaint();
   }

   private void updateJListModel(final JList pJList, final String pMethodName) {
      try {
         final Method method = studentModel.getClass().getDeclaredMethod(pMethodName);
         // Need to forcefully cast as invoke returns Object
         @SuppressWarnings("unchecked")
		final List<? extends Object> resultList = (List<? extends Object>) method.invoke(studentModel);
         final DefaultListModel newModel = new DefaultListModel();
         for (final Object loopElement : resultList) {
            newModel.addElement(loopElement);
         }
         pJList.setModel(newModel);
      } catch (Exception ex) {
         ex.printStackTrace();
         setStatus("message.exception", ex.getMessage());
      }
   }

   private List<? extends JComponent> enableComponents(final List<? extends JComponent> components, final boolean state) {
      final List<JComponent> changedComponents = new LinkedList<JComponent>();
      for (final JComponent loopComponent : components) {
         if (loopComponent.isEnabled() != state) {
            loopComponent.setEnabled(state);
            changedComponents.add(loopComponent);
         }
      }
      return changedComponents;
   }

   private JPanel createUiInputButtonsPanel() {
      final JPanel buttonsPanel = new JPanel(new GridLayout(2, 4));

      final JButton changeButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(changeButton, bundle.getString("button.change"));
      changeButton.addActionListener(new DynamicAction(this, "createChange"));
      buttonsPanel.add(changeButton);
      actionElements.add(changeButton);

      final JButton outfitButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(outfitButton, bundle.getString("button.outfit"));
      outfitButton.addActionListener(new DynamicAction(this, "createOutfit"));
      buttonsPanel.add(outfitButton);
      actionElements.add(outfitButton);

      final JButton launderButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(launderButton, bundle.getString("button.launder"));
      launderButton.addActionListener(new DynamicAction(this, "createLaunder"));
      buttonsPanel.add(launderButton);
      actionElements.add(launderButton);

      final JButton foldButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(foldButton, bundle.getString("button.fold"));
      foldButton.addActionListener(new DynamicAction(this, "createFold"));
      buttonsPanel.add(foldButton);
      actionElements.add(foldButton);

      final JButton receiveButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(receiveButton, bundle.getString("button.receive"));
      receiveButton.addActionListener(new DynamicAction(this, "createReceive"));
      buttonsPanel.add(receiveButton);
      actionElements.add(receiveButton);

      final JButton loseButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(loseButton, bundle.getString("button.lose"));
      loseButton.addActionListener(new DynamicAction(this, "createLose"));
      buttonsPanel.add(loseButton);
      actionElements.add(loseButton);

      final JButton saveCommandsButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(saveCommandsButton, bundle.getString("button.saveCommands"));
      saveCommandsButton.addActionListener(new DynamicAction(this, "saveExecutedCommands"));
      buttonsPanel.add(saveCommandsButton);
      actionElements.add(saveCommandsButton);

      final JButton exitButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(exitButton, bundle.getString("button.exit"));
      exitButton.addActionListener(new DynamicAction(Application.getInstance(), "exit"));
      buttonsPanel.add(exitButton);
      actionElements.add(exitButton);

      return buttonsPanel;
   }

   private Vector<String> getSupportedPants() {
      return new Vector<String>(Arrays.asList("black-ink-grunge", "black", "blue-chrome-rain-short",
            "blue-chrome-rain", "blue-retro", "blue-tiedyed", "blue", "crumpled-paper",
            "dark-blue-denim-short", "default", "denim-blue", "firey-orange", "glossy-black-short",
            "glossy-silver", "green-floral-shorts", "green-floral", "green", "grunge-brushed-metal",
            "pink-black-cherry-blossom", "pink", "quilted-floral-short", "quilted-floral",
            "red-and-gold-paint-splatter", "red", "vibrant-pattern", "yellow-black-striped"));
   }

   private Vector<String> getSupportedShirts() {
      return new Vector<String>(Arrays.asList("black-formal", "black", "blue-formal", "blue-retro-rusted",
            "blue-t", "blue-tiedyed", "blue", "crumpled-paper", "dark-blue-denim", "default",
            "green-checkered", "green", "orange", "pink-black-cherry-blossom", "pink", "purple",
            "quilted-floral", "red", "silver", "vibrant-pattern", "waxed-wood", "yellow-black-striped"));
   }

   private Vector<String> getSupportedSocks() {
      return new Vector<String>(Arrays.asList("addidas-white", "athletic-crew", "calzettoni-white",
            "default", "gb-black", "gizmo-pink", "green-socks", "nike-white", "rei-black", "rei-blue",
            "rei-green"));
   }

   private Map<String, Vector<String>> getGarmentAdjectivesMap() {
      final Map<String, Vector<String>> garmentAdjectivesMap = new HashMap<String, Vector<String>>();
      garmentAdjectivesMap.put(Pants.class.getSimpleName(), getSupportedPants());
      garmentAdjectivesMap.put(Shirt.class.getSimpleName(), getSupportedShirts());
      garmentAdjectivesMap.put(Socks.class.getSimpleName(), getSupportedSocks());
      return garmentAdjectivesMap;
   }

   private Map<String, Class<? extends Garment>> getGarmentTypes() {
      final Map<String, Class<? extends Garment>> garmentTypes = new TreeMap<String, Class<? extends Garment>>();
      garmentTypes.put(Pants.class.getSimpleName(), Pants.class);
      garmentTypes.put(Shirt.class.getSimpleName(), Shirt.class);
      garmentTypes.put(Socks.class.getSimpleName(), Socks.class);
      return garmentTypes;
   }

   private ActionListener getGarmentComboListener(final Map<String, Vector<String>> pGarmentAdjectivesMap, final JComboBox pGarmentAdjectiveCombo, final JComboBox pGarmentTypeCombo) {
      return new ActionListener() {
         public void actionPerformed(final ActionEvent e) {
            final String selectedGarment = (String) pGarmentTypeCombo.getSelectedItem();
            final DefaultComboBoxModel comboBoxModel = (DefaultComboBoxModel) pGarmentAdjectiveCombo.getModel();
            comboBoxModel.removeAllElements();
            for (final String loopAdj : pGarmentAdjectivesMap.get(selectedGarment)) {
               comboBoxModel.addElement(loopAdj);
            }
         }
      };
   }

   private Garment getNewGarment(final String pDialogTitleKey, final boolean pShowAdjectives) throws NoSuchMethodException, InstantiationException, IllegalAccessException, InvocationTargetException {
      final JDialog changeDialog = new JDialog(frame, bundle.getString(pDialogTitleKey), true);

      final Map<String, Class<? extends Garment>> garmentTypes = getGarmentTypes();
      final Map<String, Vector<String>> garmentAdjectivesMap = getGarmentAdjectivesMap();

      final Vector<String> garmentOptions = new Vector<String>(garmentTypes.keySet());
      final JComboBox garmentAdjectiveCombo = new JComboBox(garmentAdjectivesMap.get(garmentOptions.get(0)));
      final JComboBox garmentTypeCombo = new JComboBox(garmentOptions);
      garmentTypeCombo.addActionListener(getGarmentComboListener(garmentAdjectivesMap, garmentAdjectiveCombo, garmentTypeCombo));

      final JPanel jp = new JPanel(new GridLayout(4, 2));
      jp.add(new JLabel(""));
      jp.add(new JLabel(""));
      jp.add(new JLabel(bundle.getString("label.garmentType")));
      jp.add(garmentTypeCombo);
      if (pShowAdjectives) {
         jp.add(new JLabel(bundle.getString("label.garmentAdjective")));
         jp.add(garmentAdjectiveCombo);
      }
      jp.add(new JLabel(""));

      final JButton closeButton = new JButton();
      MnemonicHelper.configureTextAndMnemonic(closeButton, bundle.getString("button.create"));
      closeButton.addActionListener(new ActionListener() {
         public void actionPerformed(final ActionEvent e) {
            closeButton.setEnabled(false); // mark that this button was clicked
            changeDialog.setVisible(false);
         }
      });
      jp.add(closeButton);

      changeDialog.setLayout(new BorderLayout());
      changeDialog.add(new JLabel(" "), BorderLayout.NORTH);
      changeDialog.add(new JLabel(" "), BorderLayout.EAST);
      changeDialog.add(jp, BorderLayout.CENTER);
      changeDialog.add(new JLabel(" "), BorderLayout.SOUTH);
      changeDialog.add(new JLabel(" "), BorderLayout.WEST);

      changeDialog.pack();

      changeDialog.setLocationRelativeTo(frame);
      changeDialog.setVisible(true);

      if (!closeButton.isEnabled()) {
         final String selectedGarmentType = (String) garmentTypeCombo.getSelectedItem();
         final Object selectedAdjective = pShowAdjectives ? garmentAdjectiveCombo.getSelectedItem() : "";

         final Constructor<? extends Garment> decConstructor = garmentTypes.get(selectedGarmentType).getDeclaredConstructor(String.class);
         final Garment newGarment = decConstructor.newInstance(selectedAdjective);
         return newGarment;
      }
      return null;
   }

   public void createChange() {
      try {
         final Garment newGarment = getNewGarment("changeDialog.title", false);
         if (newGarment != null) {
            handleUiGeneratedCommand(new Change(newGarment));
         }
      } catch (Exception ex) {
         setStatus("message.exception", ex.getMessage());
      }
   }

   public void createReceive() {
      try {
         final Garment newGarment = getNewGarment("receiveDialog.title", true);
         if (newGarment != null) {
            handleUiGeneratedCommand(new Receive(newGarment));
         }
      } catch (Exception ex) {
         setStatus("message.exception", ex.getMessage());
      }
   }

   public void createLose() {
      try {
         final Garment newGarment = getNewGarment("loseDialog.title", true);
         if (newGarment != null) {
            handleUiGeneratedCommand(new Lose(newGarment));
         }
      } catch (Exception ex) {
         setStatus("message.exception", ex.getMessage());
      }
   }

   public void createOutfit() {
      handleUiGeneratedCommand(new Outfit());
   }

   public void createLaunder() {
      handleUiGeneratedCommand(new Launder());
   }

   public void createFold() {
      handleUiGeneratedCommand(new Fold());
   }

   public void saveExecutedCommands() {
      final JFileChooser fc = new JFileChooser(new File(System.getProperty("user.dir")));
      fc.setFileSelectionMode(JFileChooser.FILES_ONLY);

      final int returnVal = fc.showSaveDialog(frame);
      if (returnVal == JFileChooser.APPROVE_OPTION) {
         final File file = fc.getSelectedFile();
         //This is where a real application would open the file.
         setStatus("message.savingToFile", file.getAbsolutePath());
         try {
            final PrintWriter pw = new PrintWriter(new FileWriter(file));
            for (final Command loopCommand : commandList) {
               pw.print(loopCommand.toString());
               pw.println();
            }
            pw.flush();
            pw.close();
            setStatus("message.savedToFile", file.getAbsolutePath());
         } catch (Exception ex) {
            setStatus("message.exception", ex.getMessage());
         }
      } else {
         setStatus("message.saveCancelled", null);
      }
   }

   private void handleUiGeneratedCommand(final Command command) {
      if (command != null) {
         final List<? extends JComponent> changedComponents = enableComponents(this.actionElements, false);
         executeCommand(command);
         enableComponents(changedComponents, true);
      }
   }
}