package edu.rice.comp211.laundry.ui;

import edu.rice.comp211.laundry.command.Command;
import edu.rice.comp211.laundry.command.CommandVisitor;
import edu.rice.comp211.laundry.garment.Garment;
import edu.rice.comp211.laundry.io.IOProcess;
import edu.rice.comp211.laundry.iterator.ReadIteratorI;
import edu.rice.comp211.laundry.list.BiListI;
import edu.rice.comp211.laundry.list.EnumI;
//import edu.rice.comp211.laundry.list.ListIOfShirt;
import edu.rice.comp211.laundry.student.Student;

import java.io.IOException;
import java.io.PrintStream;
import java.util.*;

/**
 * Date: Mar 24, 2010
 * Time: 12:26:45 AM
 */
public class StudentModel {

   private final List<Student> students;
   private final Map<String, CommandVisitor<String>> commandVisitorMap;
   private final Map<String, IOProcess> ioProcessMap;
   private final List<Thread> runningThreads;
   private final List<String> execCommands;

   public StudentModel() {
      super();
      students = new ArrayList<Student>();
      commandVisitorMap = new HashMap<String, CommandVisitor<String>>();
      ioProcessMap = new HashMap<String, IOProcess>();
      runningThreads = new ArrayList<Thread>();
      execCommands = new Vector<String>(); // Must be vector, we want synchronization
   }

   //PrintStream pOut; 
   Map<String, PrintStream> studentPrintStrm = new Hashtable<String, PrintStream>();
   
   protected void addStudent(final Student pStudent, final IOProcess pIOProcess) {
      if (!students.isEmpty()) {
         validateStudent(students.get(0), pStudent);
      }
      students.add(pStudent);
      if (pIOProcess != null) {
         final String studentName = pStudent.name();
         PrintStream pOut = new PrintStream(pIOProcess.open(pStudent, false));
         studentPrintStrm.put(studentName, pOut);
         commandVisitorMap.put(studentName, pStudent.getCommandVisitor());//pIOProcess, false));
         ioProcessMap.put(studentName, pIOProcess);
      }
   }

   protected Student firstStudent() {
      if (students.isEmpty()) {
         throw new IllegalStateException("Student list is empty");
      }
      return students.get(0);
   }

   private void validateStudent(final Student pFirstStudent, final Student pNewStudent) {
      final String errMsg;
      if (pFirstStudent == null || pNewStudent == null) {
         errMsg = "Students may not be null";
      } else if (pFirstStudent.cleanShirts() != pNewStudent.cleanShirts()) {
         errMsg = "Students must share the same clean shirts pile";
      } else if (pFirstStudent.cleanPants() != pNewStudent.cleanPants()) {
         errMsg = "Students must share the same clean pants pile";
      } else if (pFirstStudent.cleanSocks() != pNewStudent.cleanSocks()) {
         errMsg = "Students must share the same clean socks pile";
      } else if (pFirstStudent.dirtyPile() != pNewStudent.dirtyPile()) {
         errMsg = "Students must share the same dirty pile";
      } else if (pFirstStudent.laundryRoom() != pNewStudent.laundryRoom()) {
         errMsg = "Students must share the same laundry room";
      } else {
         errMsg = null;
      }
      if (errMsg != null) {
         throw new IllegalArgumentException(errMsg);
      }
   }

   protected void executeCommand(final Command pCommand) {
      executeCommand(0, pCommand);
   }

   protected void executeCommand(final int index, final Command pCommand) {
      if (index >= 0 && index < students.size() && pCommand != null) {
         final String studentName = students.get(index).name();
         PrintStream printStrm = studentPrintStrm.get(studentName);
         printStrm.println(pCommand.accept(commandVisitorMap.get(studentName)));
         //pOut.println(pCommand.accept(commandVisitorMap.get(studentName)));
         execCommands.add(studentName + ": " + pCommand);
      }
   }

   protected List<Object> currentGarments() {
      final List<Object> garments = new ArrayList<Object>();
      for (final Student loopStudent : students) {
         garments.add(loopStudent.name());
         garments.add(loopStudent.shirt());
         garments.add(loopStudent.pants());
         garments.add(loopStudent.socks());
         garments.add("==========");
      }
      return garments;
   }

   private <T> List<T> convertToList(EnumI<T> inputEnum) {
      final List<T> result = new LinkedList<T>();

      final ReadIteratorI<T> readIter = inputEnum.newIterator();
      while (!readIter.atEnd()) {
         result.add(readIter.currentItem());
         readIter.next();
      }

      return result;
   }

   protected List<Garment> cleanShirts() {
      // We expect students to share the clean shirts shelf
      if (!students.isEmpty()) {
         return convertToList(students.get(0).cleanShirts());
      } else {
         return Collections.emptyList();
      }
   }

   protected List<Garment> cleanPants() {
      // We expect students to share the clean pants shelf
      if (!students.isEmpty()) {
         return convertToList(students.get(0).cleanPants());
      } else {
         return Collections.emptyList();
      }
   }

   protected List<Garment> cleanSocks() {
      // We expect students to share the clean socks shelf
      if (!students.isEmpty()) {
         return convertToList(students.get(0).cleanSocks());
      } else {
         return Collections.emptyList();
      }
   }

   protected List<Garment> dirtyPile() {
      // We expect students to share the dirty pile
      if (!students.isEmpty()) {
         return convertToList(students.get(0).dirtyPile());
      } else {
         return Collections.emptyList();
      }
   }

   protected List<Garment> laundryRoom() {
      // We expect students to share the dirty pile
      final List<Garment> result = new ArrayList<Garment>();
      if (!students.isEmpty()) {
         final Garment separator = null;
         final BiListI<BiListI<Garment>> laundryPiles = students.get(0).laundryRoom();
         final ReadIteratorI<BiListI<Garment>> readIter = laundryPiles.newIterator();

         while (!readIter.atEnd()) {
            result.addAll(convertToList(readIter.currentItem()));
            result.add(separator);
            readIter.next();
         }
      }

      return result;
   }

   public List<String> getExecCommands() {
      return execCommands;
   }

   protected synchronized void clearStudents() {
      students.clear();
      commandVisitorMap.clear();
      ioProcessMap.clear();
      runningThreads.clear();
      execCommands.clear();
   }
   
   protected synchronized boolean doneExecuting() {
      for (final Thread loopThread : runningThreads) {
         if (!Thread.State.TERMINATED.equals(loopThread.getState())) {
            return false;
         }
      }
      if (!runningThreads.isEmpty()) {
         runningThreads.clear();
      }
      return true;
   }

   protected synchronized void runOneStep() {
      for (final Student loopStudent : students) {
         final Thread loopThread = new Thread(new Runnable() {
            public void run() {
               try {
                  final IOProcess ioProcForStud = ioProcessMap.get(loopStudent.name());
                  final Command nextCommand = ioProcForStud.nextCommand();
                  if (nextCommand != null) {
                     final String studentName = loopStudent.name();
                     PrintStream printStrm = studentPrintStrm.get(studentName);
                     printStrm.println(studentName+": "+nextCommand.accept(commandVisitorMap.get(studentName)));
                     execCommands.add(studentName + ": " + nextCommand);
                  }
               } catch (final IOException e) {
                  e.printStackTrace();
               }
            }
         });
         runningThreads.add(loopThread);
      }
      for (final Thread loopThread : runningThreads) {
         if (Thread.State.NEW.equals(loopThread.getState())) {
            loopThread.start();
         }
      }
   }

   protected synchronized void runAllSteps() {
      for (final Student loopStudent : students) {
         final Thread loopThread = new Thread(new Runnable() {
            public void run() {
               try {
                  final IOProcess ioProcForStud = ioProcessMap.get(loopStudent.name());
                  Command nextCommand = ioProcForStud.nextCommand();
                  while (nextCommand != null) {
                     final String studentName = loopStudent.name();
                     PrintStream printStrm = studentPrintStrm.get(studentName);
                     printStrm.println(studentName+": "+nextCommand.accept(commandVisitorMap.get(studentName)));
                     execCommands.add(studentName + ": " + nextCommand);
                     Thread.sleep(((long)(Math.random() * 10000)) % 300);
                     nextCommand = ioProcForStud.nextCommand();
                  }
               } catch (final Exception ex) {
                  ex.printStackTrace();
               }
            }
         });
         runningThreads.add(loopThread);
      }
      for (final Thread loopThread : runningThreads) {
         if (Thread.State.NEW.equals(loopThread.getState())) {
            loopThread.start();
         }
      }
   }
}
